#include <iostream>
#include <stdio.h>
#include <valarray>
#include <map>
#include "mydebug.h"

#include "optics.h"
#include "complexfiles.h"
#include "configfiles.h"
#include "cubicspline.h"
#include "fourier.h"

using namespace std;

/*! 
  @function CalculateE_Omega_K_z_ForDiracInfiniteMedium
  @abstract Calculate E spectral density for a thin polarization sheet in an infinite medium
  @discussion See JOSA B article, Eq.(x)
  @param inK An array of real values containing the values of the transverse wave number K where the 
             field  will be calculated
  @param inOmega Frequency for which the calculation is done
  @param inEpsilon1 Complex dielectric constant of the medium
  @result A complex val_array containing the spectral density at all the K points provided in inK
*/ 


valarray<Complex> CalculateE_Omega_K_z_ForDiracInfiniteMedium (	
								const valarray<double>& inK,
								double inOmega,
								Complex inEpsilon1,
								const valarray<Complex>& inPol_array);

/*! 
  @function CalculateE_Omega_K_z_ForThreeLayerGeometry
  @abstract Calculate E spectral density for a thin polarization sheet in an infinite medium
  @discussion See JOSA B article, Eq.(x)
  @param inK An array of real values containing the values of the transverse wave number K where the 
             field  will be calculated
  @param inOmega Frequency for which the calculation is done
  @param inEpsilon1 Complex dielectric constant of the medium 1
  @param inEpsilon2 Complex dielectric constant of the medium 2
  @param inEpsilon3 Complex dielectric constant of the medium 3
  @result A complex val_array containing the spectral density at all the K points provided in inK
*/ 

valarray<Complex> CalculateE_Omega_K_z_ForThreeLayerGeometry (	
								const valarray<double>& inK,
								double inOmega,
								double inNg,
								double inAlpha,
								double inAlpha_transverse,
								double inD,
								Complex inEpsilon1,
								Complex inEpsilon2,
								Complex inEpsilon3,
								const valarray<Complex>& inPol_array,
								double phaseMismatchFactor);

void
ApplyLensTransform(const valarray<double>& inK, valarray<Complex>& ioVx, double Omega,  Complex e1, double f, double L, double d_o, double d_i);

void
ApplyPropagationTransform(const valarray<double>& inK, valarray<Complex>& ioVx, double Omega,  Complex e1, double d_o);

Complex
IntegrateKdKover4pi2_WRT_arbK( const valarray<double>& inK,
					const valarray<Complex>& inV,

                                        #ifdef HAVE_NUMERIC_LIMITS
                                            const double nu = numeric_limits<double>::infinity()
                                        #else
                                            const double nu = 1e100
                                        #endif
                                        );

void
InitPeakPolarizationFromFile(string filename, 
                         valarray<double>& ioOmega_array, 
                         valarray<Complex>& ioPeakPolarizationFromFile);
                         
valarray<Complex> PeakPolarizationFromFile(	const valarray<double>&  inK,
		double inK_BW,
		long inOmegaIndex,
		valarray<Complex>& inPeakPolarizationFromFile);
		
valarray<Complex> PeakPolOneColourRect(	const valarray<double>&  inK,
		double inK_BW,
		double inOmega,
		double inOmega_BW,
		double inTheoristsEwInCrystal,
		double inChi_eff);
		
valarray<Complex> PeakPolOneColourShift(	const valarray<double>&  inK,
		double inK_BW,
		double inOmega,
		double inOmega_BW,
		double inTheoristsE2wInCrystal,
		double inChi_eff,
		double inTau_eff,
		double inWp_eff);

valarray<Complex> PeakPolTwoColourInj(	const valarray<double>&  inK,
		double inK_BW,
		double inOmega,
		double inOmega_BW,
		double inTheoristsEwInCrystal,
		double inTheoristsE2wInCrystal,
		double inChi_eff,
		double inTau_eff,
		double inWp_eff);

Complex EpsilonAtOmegaForMaterial(double inOmega, unsigned long inMaterial);
Complex EOFilterAtOmega(double inOmega, Complex inEpsilon, double l, double ng_pump );

enum {kInvalid 				= 1000,
	kGaAs					= 1,
	kZnTe					= 2,
	kGaP					= 3, 
	kAir					= 4, 
	kGaAsFrequencyDependent	= 5,
    kZnTeFrequencyDependent	= 6, 
    kGaPFrequencyDependent	= 7,
    kZnTeFrequencyDependentUpTo5THz = 8};
    
enum {kOneColourRectification=1, kOneColourShift, kOneColourInjection, kTwoColourRectification, kTwoColourShift, kTwoColourInjection, kPolarizationFromFile};
enum {kInfiniteMedium=1, kThreeLayerStructure};

Complex gEpsilon[kAir+1];
cubicspline gGaAs_n_FrequencyDependent;
cubicspline gZnTe_n_FrequencyDependent;
cubicspline gGaP_n_FrequencyDependent;
cubicspline gGaAs_k_FrequencyDependent;
cubicspline gZnTe_k_FrequencyDependent;
cubicspline gGaP_k_FrequencyDependent;


	#if 0
		ifstream IN("parameters.txt");
		ofstream OUT("out.txt");
	#else
		#define IN cin
		#define OUT cout
	#endif


int
main()
{

	
//	StProfileSection p("\pProfiler",100,100);
	
	try {

	// Variable definition
	
	double Omega = NAN, dOmega= NAN, T_Max= NAN, dT= NAN;
	double dKo= NAN;
	double D= NAN, ng= NAN, alpha_current= NAN, alpha_transverse=NAN, theoristsEwInCrystal= NAN, theoristsE2wInCrystal= NAN;
	double Ew= NAN, E2w= NAN, chi_eff= NAN,tau_eff= NAN,wp_eff= NAN;
	double f= NAN,L= NAN,d_o= NAN,d_i= NAN;
	double c = 3e8,EOFilter_L,EOFilter_ng_probe;
	complex_double epsilon1, epsilon2, epsilon3, epsilonEOFilter;
	complex_double F_s, F_p, L_plus, L_minus;
	complex_double w1, w2, w3, nu1, nu2, nu3;
	unsigned long material1=kInvalid,material2=kInvalid,material3=kInvalid,EOFilter_material=kInvalid;
	unsigned long process=0, geometry=0;
	bool withlens, farFieldPropagation, withEOFilter;
	string filename;
	
	unsigned long N_Omega=0, N_K=0;
	double Omega_Max = NAN, Omega_BW = NAN, K_Max = NAN, K_BW = NAN, phaseMismatchFactor = NAN;

	valarray<double> Omega_array;
	valarray<double> Frequency_array;
	valarray<Complex> spectrum;
	valarray<Complex> current;
	valarray<Complex> Integrand;
	valarray<Complex> Pol_array;
	valarray<Complex> eofilter;
	valarray<Complex>  peakPolarizationFromFile;

	map<string, string> varDict;
	
	ReadParametersFromStream(IN, varDict);
	
	InitVariable(varDict, "process",process);

	InitVariable(varDict, "N_K",N_K);
	InitVariable(varDict, "K_Max",K_Max);
	InitVariable(varDict, "K_BW",K_BW);
	InitVariable(varDict, "d_o",d_o);
	InitVariable(varDict, "geometry",geometry);
	InitVariable(varDict, "alpha_current",alpha_current);

	if (process == kPolarizationFromFile) {
		InitVariable(varDict, "peakPolarizationFile",filename);
		InitPeakPolarizationFromFile(filename, Omega_array, peakPolarizationFromFile);
	} else {
		InitVariable(varDict, "N_Omega",N_Omega);
		InitVariable(varDict, "Omega_Max",Omega_Max);
		InitVariable(varDict, "Omega_BW",Omega_BW);
		InitVariable(varDict, "Ew",Ew);
		InitVariable(varDict, "chi_eff",chi_eff);

        dOmega = double(2.) * Omega_Max / N_Omega;
        Omega_array.resize(N_Omega);
        for (unsigned long i = 0; i < Omega_array.size()/2 ; i++) {
        	Omega_array[i] = double(i) * dOmega;
        }

	}

    Frequency_array.resize(Omega_array.size());
    spectrum.resize(Omega_array.size());
    current.resize(Omega_array.size());
    eofilter.resize(Omega_array.size());

	if (VariableExists(varDict, "phaseMismatchFactor")) {
		InitVariable(varDict, "phaseMismatchFactor",phaseMismatchFactor);
	} else {
		phaseMismatchFactor = 1.;
	}

	if (geometry == kThreeLayerStructure) {
		InitVariable(varDict, "ng",ng);
		InitVariable(varDict, "D",D);
		InitVariable(varDict, "material1",material1);
		InitVariable(varDict, "material2",material2);
		InitVariable(varDict, "material3",material3);
		if (VariableExists(varDict, "alpha_transverse")) {
			InitVariable(varDict, "alpha_transverse",alpha_transverse);
		} else {
			alpha_transverse = 0.;
		}

	} else if (geometry == kInfiniteMedium) {
		InitVariable(varDict, "material1",material1);
		material2 = kInvalid;
		material3 = kInvalid;
	}



	if (VariableExists(varDict, "withlens")) {
		withlens = true;
		InitVariable(varDict, "f",f);
		InitVariable(varDict, "L",L);
		InitVariable(varDict, "d_i",d_i);
	} else {
		withlens = false;
	}
	if (VariableExists(varDict, "farFieldPropagation")) {
		farFieldPropagation = true;
	} else {
		farFieldPropagation = false;
	}

	if (VariableExists(varDict, "withEOFilter")) {
		withEOFilter = true;
		InitVariable(varDict, "EOFilter_material",EOFilter_material);
		InitVariable(varDict, "EOFilter_L",EOFilter_L);
		InitVariable(varDict, "EOFilter_ng_probe",EOFilter_ng_probe);
	} else {
		withEOFilter = false;
	}
	
	if (process == kOneColourShift || process == kOneColourInjection || process == kTwoColourShift || process == kTwoColourInjection) {
		InitVariable(varDict, "E2w",E2w);
		InitVariable(varDict, "tau_eff",tau_eff);
		InitVariable(varDict, "wp_eff",wp_eff);
	}


	dKo = K_Max / N_K;
	
	// Assumes a reflection of the pump from index of 3.3
	theoristsEwInCrystal = 0.44 * Ew / 2;
	theoristsE2wInCrystal = 0.44 * E2w / 2;
	

	ofstream file("gnuplot.dem");
	file << "set terminal moov\n";

	

	size_t recommendedKSize = 2 * N_K;
	for (unsigned long i = 0; i < Omega_array.size()/2 ; i++) {


		if (! farFieldPropagation) { 
			/* NEAR field propagation algorithm */

			/*	
			 Determination of sampling in K
			*/

			Omega = Omega_array[i];
			
			if (geometry == kThreeLayerStructure) {
				epsilon1 = EpsilonAtOmegaForMaterial(Omega, material1);
				epsilon2 = EpsilonAtOmegaForMaterial(Omega, material2);
				epsilon3 = EpsilonAtOmegaForMaterial(Omega, material3);
			} else if (geometry == kInfiniteMedium) {
				epsilon1 = EpsilonAtOmegaForMaterial(Omega, material1);
				
			}

			valarray<double> K_array;

			size_t arraysize = recommendedKSize / 2;
			unsigned long j;
			
			if (i % 10 == 0)
				OUT << "Omega = " << Omega << "[" << i << "/" << Omega_array.size() << "]\n";

	 		OUT << "Attempting K array with ";
			do {
				K_array.resize(arraysize, 0.);
				arraysize *= 2;

				for (j = 0; j < K_array.size(); j++) {
					double Delta_K;
						
					if (j == 0) {
						K_array[j] = 0.;
						Delta_K = dKo;
					} else {
						K_array[j] = K_array[j-1] + Delta_K;

						double wo;
						
						if (geometry == kThreeLayerStructure) {
							wo = abs(w(Omega, K_array[j],epsilon3 ));
						} else {
							wo = abs(w(Omega, K_array[j],epsilon1 ));
						}
						
						Delta_K =  0.1 * wo / K_array[j] * dKo ;

					}
					
					if (Delta_K > dKo) {
						Delta_K = dKo;
					} else if (Delta_K == 0) {
						Delta_K = dKo;
					}
				}
			OUT << K_array.size() << " elements...";

			} while (K_array[j-1]/K_Max < 0.95);
			OUT << "Done\n";
			recommendedKSize = K_array.size();
			Integrand.resize(recommendedKSize);
			Pol_array.resize(recommendedKSize);
			/*
			 Polarization source
			*/
			switch (process) {
				case kPolarizationFromFile:				
					Pol_array = PeakPolarizationFromFile(K_array, K_BW, i, peakPolarizationFromFile);
					break;
				case kOneColourRectification:
					
					Pol_array = PeakPolOneColourRect(K_array, K_BW, Omega, Omega_BW, theoristsEwInCrystal, chi_eff);
					break;
				case kOneColourShift:
					Pol_array = PeakPolOneColourShift(K_array, K_BW, Omega, Omega_BW, theoristsE2wInCrystal, chi_eff, tau_eff,wp_eff);
					break;
				case kOneColourInjection:
				case kTwoColourRectification:
				case kTwoColourShift:
					OUT << "Polarization source " << process << " is not supported.";
					exit(1);
					break;
				case kTwoColourInjection:
					Pol_array = PeakPolTwoColourInj(K_array, K_BW, Omega, Omega_BW, theoristsEwInCrystal, theoristsE2wInCrystal, chi_eff, tau_eff,wp_eff);
					break;
			}
			// Also calculate the current at R=0
			current[i] =  SQR(K_BW/2./sqrt(pi)) * (-I * Omega) * Pol_array[0];

			/*
			 Geometry of the structure
			*/
			switch (geometry) {
				case kInfiniteMedium:
					Integrand = CalculateE_Omega_K_z_ForDiracInfiniteMedium(K_array, Omega, epsilon1, Pol_array);
					break;
				case kThreeLayerStructure:
					Integrand = CalculateE_Omega_K_z_ForThreeLayerGeometry(K_array, Omega, ng,alpha_current, alpha_transverse, D, epsilon1, epsilon2, epsilon3, Pol_array, phaseMismatchFactor);
					break;
				default:
					OUT << "Geometry " << geometry << " is not supported.";
					exit(1);
					break;
					
			}
			
			/* Integral over phi is just integral of d_phi */
			
			Integrand *= 2. * pi;
			
			/*
			 Other optical transformations
			*/
			if (withlens) {
				ApplyLensTransform(K_array, Integrand, Omega, epsilon1, f, L, d_o, d_i);

				spectrum[i] =  IntegrateKdKover4pi2_WRT_arbK( K_array, Integrand);
				spectrum[i] *= SQR(erf(Omega/c*L/f/K_BW));
			} else {
				ApplyPropagationTransform(K_array, Integrand, Omega, epsilon1, d_o);
				spectrum[i] =  IntegrateKdKover4pi2_WRT_arbK( K_array, Integrand);
			}		

			/*
			 Other filter functions
			*/
			
			if (withEOFilter) {
				eofilter[i] = EOFilterAtOmega(Omega, EpsilonAtOmegaForMaterial(Omega, EOFilter_material), EOFilter_L, EOFilter_ng_probe);
				spectrum[i] *= eofilter[i];				
			}
		
		} else {
			/* FAR field propagation algorithm */

			/*
			   0. Setup variables
			*/
			Omega = i * dOmega;	
			Omega_array[i] = Omega;

			if (geometry == kThreeLayerStructure) {
				epsilon1 = EpsilonAtOmegaForMaterial(Omega, material1);
				epsilon2 = EpsilonAtOmegaForMaterial(Omega, material2);
				epsilon3 = EpsilonAtOmegaForMaterial(Omega, material3);
			} else if (geometry == kInfiniteMedium) {
				epsilon1 = EpsilonAtOmegaForMaterial(Omega, material1);
				
			}

			valarray<double> K_array(1);
			
			K_array[0] = 0.;

			valarray<Complex> Pol_array;

			/*
			 Polarization source
			*/
			
			switch (process) {
				case kPolarizationFromFile:				
					Pol_array = PeakPolarizationFromFile(K_array, K_BW, i, peakPolarizationFromFile);
					break;
				case kOneColourRectification:
					
					Pol_array = PeakPolOneColourRect(K_array, K_BW, Omega, Omega_BW, theoristsEwInCrystal, chi_eff);
					current[i] = SQR(K_BW/2/sqrt(pi)) * (-I * Omega) * Pol_array[0];
					break;
				case kOneColourShift:
					Pol_array = PeakPolOneColourShift(K_array, K_BW, Omega, Omega_BW, theoristsE2wInCrystal, chi_eff, tau_eff,wp_eff);
					current[i] = SQR(K_BW/2/sqrt(pi)) * (-I * Omega) * Pol_array[0];
					break;
				case kOneColourInjection:
				case kTwoColourRectification:
				case kTwoColourShift:
					OUT << "Polarization source " << process << " is not supported.";
					exit(1);
					break;
				case kTwoColourInjection:
					Pol_array = PeakPolTwoColourInj(K_array, K_BW, Omega, Omega_BW, theoristsEwInCrystal, theoristsE2wInCrystal, chi_eff, tau_eff,wp_eff);
					// Also calculate the current at K=0
					current[i] = SQR(K_BW/2/sqrt(pi)) * (-I * Omega) * Pol_array[0];
					break;
			}
			CheckArray_(Pol_array);

			/*
			 Geometry of the structure
			*/

			switch (geometry) {
				case kInfiniteMedium:
					Integrand = CalculateE_Omega_K_z_ForDiracInfiniteMedium(K_array, Omega, epsilon1, Pol_array);
					break;
				case kThreeLayerStructure:
					Integrand = CalculateE_Omega_K_z_ForThreeLayerGeometry(K_array, Omega, ng,alpha_current, alpha_transverse, D, epsilon1, epsilon2, epsilon3, Pol_array, phaseMismatchFactor);
					break;
				default:
					OUT << "Geometry " << geometry << " is not supported.";
					exit(1);
					break;
					
			}
			
			CheckArray_(Integrand);

			/*
			 Other optical transformations
			*/
			
			spectrum[i] =  Integrand[0] * (-I * w(Omega, K_array[0], epsilon1) / Complex(2. * pi)  / Complex(d_o)) ;

			/*
			 Other filter functions
			*/
			
			if (withEOFilter) {
				eofilter[i] = EOFilterAtOmega(Omega, EpsilonAtOmegaForMaterial(Omega, EOFilter_material), EOFilter_L, EOFilter_ng_probe);
				spectrum[i] *= eofilter[i];
			}
			

			CheckArray_(spectrum);
		
		}			
	}

	CheckArray_(spectrum);
	for (unsigned long i = Omega_array.size() - 1; i >= Omega_array.size()/2; i--) {
		unsigned long j = Omega_array.size() - i;
		Omega_array[i] = -Omega_array[j];
		eofilter[i] = conj(eofilter[j]);
		spectrum[i] = conj(spectrum[j]);
		current[i] = conj(current[j]);
	}
	CheckArray_(spectrum);

	file << "set terminal pict\n";

	Frequency_array = Omega_array/2./pi;
	WriteComplexFunctionsToDisk(Frequency_array,spectrum,"spectrum_abs.dat","spectrum_arg.dat");
	WriteComplexFunctionsToDisk(Frequency_array,current,"current_abs.dat","current_arg.dat");
	WriteComplexFunctionsToDisk(Frequency_array,eofilter,"eofilter_abs.dat","eofilter_arg.dat");
	
	WriteRealFFTToDisk(Frequency_array,spectrum,-1,"temporal.dat");
	WriteRealFFTToDisk(Frequency_array,current,-1,"temporal_current.dat");


	// Debug output, just to check values
	OUT << "\n\nValues were:" << endl;
	OUT << "Process " << process << endl;
	OUT << "Geometry " << geometry << endl;
	OUT << "Number of points in Omega " << Omega_array.size() << endl;
	OUT << "Maximum Omega [2 Pi x Hz ] :" << Omega_Max << endl;
	OUT << "Omega bandwidth [2 Pi x Hz ] :" << Omega_BW << endl;
	OUT << "Number of points in K " << N_K<< endl;
	OUT << "Maximum K [2 Pi x m^(-1} ] :" <<  K_Max<< endl;
	OUT << "K bandwidth [2 Pi x m^(-1} ] :" << K_BW<< endl;
	OUT << "Material thickness D [m] :" << D<< endl;
	OUT << "Pump group index :" << ng<< endl;
	OUT << "Current alpha :" << alpha_current << endl;
	OUT << "Electric field omega (outside crystal) [V/m]:" << Ew<< endl;
	OUT << "Electric field 2*omega (outside crystal) [V/m]:" << E2w<< endl;
	OUT << "Nonlinear chi_eff :" << chi_eff<< endl;
	OUT << "tau_relax effective:" << tau_eff<< endl;
	OUT << "omega_p effective :" << wp_eff<< endl;
	
	OUT << "Material epsilon1 ("<<kGaAs << "=GaAs, "<< kZnTe<<"=ZnTe, " << kGaP << "= GaP, " << kAir << "=air)" << material1<< endl;
	OUT << "Material epsilon2 ("<<kGaAs << "=GaAs, "<< kZnTe<<"=ZnTe, " << kGaP << "= GaP, " << kAir << "=air)"<< material2<< endl;
	OUT << "Material epsilon3 ("<<kGaAs << "=GaAs, "<< kZnTe<<"=ZnTe, " << kGaP << "= GaP, " << kAir << "=air)"<<  material3<< endl;
	OUT << "Distance where profile is measured:" << d_o<< endl;
	OUT << endl;

	} catch (exception e) {
	
		cerr << "Caught exception.  Program terminated (" << e.what() << ")\n" ;
	}
	return 0;	
}

valarray<Complex> CalculateE_Omega_K_z_ForDiracInfiniteMedium (	
								const valarray<double>& inK,
								double inOmega,
								Complex inEpsilon1,
								const valarray<Complex>& inPol_array)
{
	valarray<Complex> Integrand(inK.size());

	double Omega = inOmega;
	Complex e1 = inEpsilon1;
	Complex nu1 = Omega/c*index(e1);
	

	CheckValue_(e1);

	
	for ( unsigned long i = 0; i < inK.size(); ++i) {
		double K =inK[i];
		Complex F_p;
		Complex F_s;
		Complex Lplus;
		Complex Lminus;
		Complex integralpluswave;
		Complex integralminuswave;
 
		Complex w1 = w(Omega, K, e1);

		Complex Pol = inPol_array[i];
		
			 
		if (Omega == 0 ) {
			integralpluswave = 0.;
			integralminuswave = 0.;
			Lplus  = 0.;
			Lminus  = 0.;
		} else {
			integralpluswave  =  1. / (2. * epsilon_o)
									* I  
									* 1. / w1 * ( 0.5 * SQR(Omega/c) + 0.5 * SQR(w1/e1) )
									* Pol;
									

		} 

		Integrand[i] =  (integralpluswave);

	}


	CheckArray_(Integrand); 
	return Integrand;
}


valarray<Complex> CalculateE_Omega_K_z_ForThreeLayerGeometry (	
								const valarray<double>& inK,
								double inOmega,
								double inNg,
								double inAlphaPolarizationOrCurrent,
								double inAlpha_transverse,
								double inD,
								Complex inEpsilon1,
								Complex inEpsilon2,
								Complex inEpsilon3,
								const valarray<Complex>& inPol_array,
								double phaseMismatchFactor)
{
	valarray<Complex> Integrand(inK.size());

	double Omega = inOmega;
	double ng = inNg;
	double alpha = inAlphaPolarizationOrCurrent;
	double alpha_transverse = inAlpha_transverse;
	double D = inD;
	Complex e1 = inEpsilon1;
	Complex e2 = inEpsilon2;
	Complex e3 = inEpsilon3;
	Complex nu1 = Omega/c*index(e1);
	Complex nu3 = Omega/c*index(e3);
	

	CheckValue_(e1);
	CheckValue_(e2);
	CheckValue_(e3);
	
	for ( unsigned long i = 0; i < inK.size(); ++i) {
		double K =inK[i];
		Complex F_p;
		Complex F_s;
		Complex Lplus;
		Complex Lminus;
		Complex integralpluswave;
		Complex integralminuswave;
 
		Complex w1 = w(Omega, K, e1);
		Complex w2 = w(Omega, K, e2);
		Complex w3 = w(Omega, K, e3);


		Complex Pol = inPol_array[i];
		
		F_p = tpij(Omega,K,e3,e1) / (1. - rpij(Omega,K,e3,e1) * rpij(Omega,K,e3,e2) * exp(2.*I*w3*D) * exp(-2.*alpha_transverse*(K/real(w3)*D)));
		F_s = tsij(Omega,K,e3,e1) / (1. - rsij(Omega,K,e3,e1) * rsij(Omega,K,e3,e2) * exp(2.*I*w3*D) * exp(-2.*alpha_transverse*(K/real(w3)*D))) ;

			 
		if (Omega == 0 ) {
			integralpluswave = 0.;
			integralminuswave = 0.;
			Lplus  = 0.;
			Lminus  = 0.;
		} else {
			Lplus = 				(exp((I*ng*Omega/c - alpha)*D) 
									* (1. - exp( -D * (I*ng*Omega/c - alpha - I*w3) )) 
									/ (I*ng*Omega/c - alpha - I*w3) );
			
			Lminus = 				(exp((I*ng*Omega/c - alpha)*D) 
									* (1. - exp( -D * (I*ng*Omega/c - alpha + I*w3) )) 
									/ (I*ng*Omega/c - alpha + I*w3) );
															
			integralpluswave  =  	1. / (4. * pi * epsilon_o)
									* 2. * pi * I  
									* 1. / w3 * (F_s * 0.5 * SQR(Omega/c) 
									          + F_p * w1*w3/(2.*index(e1)*index(e3)) ) 
									* Pol
									* Lplus;
									
			integralminuswave = 	exp(2.*I*w3*D) 
									* 1. / (4. * pi * epsilon_o) 
									* 2. * pi * I 
									* 1. / w3 * (rsij(Omega,K,e3,e2) * F_s * 0.5 * SQR(Omega/c) 
									          - rpij(Omega,K,e3,e2) * F_p * w1*w3/(2.*index(e1)*index(e3)) ) 
									* Pol
									* Lminus;

		} 

		Integrand[i] = phaseMismatchFactor * (integralpluswave + integralminuswave);


	}

	CheckArray_(Integrand); 
	return Integrand;
}

void
ApplyLensTransform(const valarray<double>& inK, valarray<Complex>& ioVx, double Omega,  Complex e1, double f, double L, double d_o, double d_i)
{
	if ( 2. * f == d_o &&  2. * f == d_i) {
		CheckDoubleValue_(Omega);
		CheckDoubleValue_(f);
		CheckDoubleValue_(L);
		CheckDoubleValue_(d_o);
		CheckDoubleValue_(d_i);                                        
		CheckValue_(e1);
		
		for ( unsigned long i = 0; i < inK.size(); ++i) {
			double K =inK[i];
	 		CheckDoubleValue_(K);

			Complex w1 = w(Omega, K, e1);
			CheckValue_(w1);

			Complex theta, theta1, theta2;
			double theta_r, theta_i;
			
			if (Omega == 0) {
				ioVx[i] = 0. ;
			} else {
				theta1 = w1 * (d_o + d_i) - Omega/c * real(index(e1)) * (d_o+d_i); 
				theta2 = 2. * f * SQR(K) / (Omega/c);
				theta = theta1 + theta2;
				theta_r = real(theta1 + theta2);
				theta_i = imag(theta1 + theta2);
	
				ioVx[i] *=  exp(-theta_i) * (cos(theta_r) + I * sin(theta_r));
//				ioVx[i] *=   exp(I*(theta) ) ;
//				OUT << theta << endl;
			}
			
			CheckValue_( theta );
		}

		CheckArray_(ioVx); 
		
	} else {
		cerr << "Can't transform for cases other than collimate-focus geometries" << endl;
		throw runtime_error("lens transform invalid");
	
	}

}

void
ApplyPropagationTransform(const valarray<double>& inK, valarray<Complex>& ioVx, double Omega,  Complex e1, double d_o)
{
	CheckDoubleValue_(Omega);
	CheckDoubleValue_(d_o);
	CheckValue_(e1);
	
	for ( unsigned long i = 0; i < inK.size(); ++i) {
		double K =inK[i];
 		CheckDoubleValue_(K);

		Complex w1 = w(Omega, K, e1);
		CheckValue_(w1);

		Complex theta;
		theta = w1 * (d_o) - Omega/c * real(index(e1)) * d_o; 
		ioVx[i] *=  exp(I*(theta) ) ;

	}

	CheckArray_(ioVx); 
}

void
InitPeakPolarizationFromFile(string filename, 
                         valarray<double>& ioOmega_array, 
                         valarray<complex_double>& ioPeakPolarizationFromFile)
{
	OUT << "We expect a temporal profile: time [tab] current density[SI units].\n";
	OUT << "The points in time must be equally separated and the number of\n";
	OUT << "points must be a power of 2 due to the algorithm used for FFT\n";	
	cubicspline temp(filename.c_str());
	valarray<double> t(temp.size());
	valarray<complex_double> current(temp.size());
	double dT, TMax, dF, FMax;
	
	for (long i = 0; i < temp.size(); i++) {
		t[i] = temp.xelement(i) ;
		current[i] = temp.yelement(i);
//		OUT << t[i] << " " << current[i]<< endl;
	}
	ExtractFFTParams(t, dT, TMax, dF, FMax);
	OUT <<  dT << " " << TMax << " " << dF << " " << FMax << endl;
	if ( FFT(current, +1) ) {
		OUT << "Error evaluating FFT";
		exit(1);
	}
//	current = current.cshift( int(current.size()/2) );
	current *= dT;
	ioOmega_array.resize(temp.size());
	ioPeakPolarizationFromFile.resize(temp.size());
	
	for (long i = 0; i < ioPeakPolarizationFromFile.size() / 2 ; i++) {
		ioOmega_array[i] = 2. * pi * dF * i ;
		if (ioOmega_array[i] != 0.) {
			ioPeakPolarizationFromFile[i] = current[i] / (- I * ioOmega_array[i]);
		} else {
			ioPeakPolarizationFromFile[i] = 0.;
		}
//		OUT << ioOmega_array[i] << " " << ioPeakPolarizationFromFile[i] << endl;
	}

	for (unsigned long i = ioOmega_array.size() - 1; i >= ioOmega_array.size()/2; i--) {
		unsigned long j = ioOmega_array.size() - i;
		ioOmega_array[i] = -ioOmega_array[j];
	}

//	WriteComplexFunctionsToDisk(ioOmega_array, current, "test_abs.dat","test_arg.dat");
	
}

valarray<Complex> PeakPolarizationFromFile(	const valarray<double>&  inK,
		double inK_BW,
		long inOmegaIndex,
		valarray<complex_double>& inPeakPolarizationFromFile)
{

	double K_BW = inK_BW;
	valarray<Complex> Pol_array(inK.size());

	complex_double peakPol = inPeakPolarizationFromFile[inOmegaIndex];
	
	for( unsigned long i = 0; i < inK.size();  i++) {
		Pol_array[i] =  peakPol * exp(- SQR(inK[i] / K_BW) )
					  * 2. * sqrt(pi) / (K_BW)
					  * 2. * sqrt(pi) / (K_BW);
	}

	return Pol_array;

}


valarray<Complex>
PeakPolOneColourRect(	const valarray<double>& inK,
		double inK_BW,
		double inOmega,
		double inOmega_BW,
		double inTheoristsEwInCrystal,
		double inChi_eff)
{
		// Electric field in inside crystal after fresnel losses.
		// Because of the definition  of Eq. (1), the experimental field
		// is two times the theorist field
		// chi_eff must take into account the projection of the tensor 

	double Ew = inTheoristsEwInCrystal;
	double Omega = inOmega;
	double Omega_BW = inOmega_BW;
	double K_BW = inK_BW;
	double chi_tilde = inChi_eff;
	valarray<Complex> Pol_array(inK.size());
	
	for( unsigned long i = 0; i < inK.size();  i++) {
		/* Eq. (42) */
		Pol_array[i] = (2 * epsilon_o * chi_tilde * Ew * Ew )
					  * exp(- SQR(Omega/ Omega_BW) )
					  * exp(- SQR(inK[i] / K_BW) )
					  * 2. * sqrt(pi) / Omega_BW 
					  * 2. * sqrt(pi) / (K_BW)
					  * 2. * sqrt(pi) / (K_BW);;
	}

	return Pol_array;
}

valarray<Complex>
PeakPolOneColourShift(	const valarray<double>& inK,
		double inK_BW,
		double inOmega,
		double inOmega_BW,
		double inTheoristsE2wInCrystal,
		double inChi_eff,
		double inTau_eff,
		double inWp_eff)
{
		// Electric field is inside crystal after fresnel losses.
		// Because of the definition  of Eq. (5), the experimental field
		// is two times the theorist field
		// chi_eff must take into account the projection of the tensor 

	double E2w = inTheoristsE2wInCrystal;
	double sigma_xxx = inChi_eff;
	double omega_p = inWp_eff;
	double tau = inTau_eff;
	double Omega = inOmega;
	double Omega_BW = inOmega_BW;
	double K_BW = inK_BW;

	valarray<Complex> Pol_array(inK.size());
	
	for( unsigned long i = 0; i < inK.size();  i++) {
		Pol_array[i] = (-I * Omega + 1. / tau) * ( 2 * sigma_xxx *  E2w *  E2w )
						/ (SQR(omega_p) + SQR(-I*Omega) - I*Omega/tau)
					  * exp(- SQR(Omega/ Omega_BW) )
					  * exp(- SQR(inK[i] / K_BW) )
					  * 2. * sqrt(pi) / Omega_BW 
					  * 2. * sqrt(pi) / (K_BW)
					  * 2. * sqrt(pi) / (K_BW);
	}

	return Pol_array;
}



valarray<Complex>
PeakPolTwoColourInj(	const valarray<double>& inK,
		double inK_BW,
		double inOmega,
		double inOmega_BW,
		double inTheoristsEwInCrystal,
		double inTheoristsE2wInCrystal,
		double inChi_eff,
		double inTau_eff,
		double inWp_eff)
{
		// Electric field is inside crystal after fresnel losses.
		// Because of the definition  of Eq. (5), the experimental field
		// is two times the theorist field
		// chi_eff must take into account the projection of the tensor 

	double Ew = inTheoristsEwInCrystal;
	double E2w = inTheoristsE2wInCrystal;
	double eta_xxxx = inChi_eff;
	double omega_p = inWp_eff;
	double tau = inTau_eff;
	double Omega = inOmega;
	double Omega_BW = inOmega_BW;
	double K_BW = inK_BW;

	valarray<Complex> Pol_array(inK.size());
	
	for( unsigned long i = 0; i < inK.size();  i++) {
		Pol_array[i] =( 2. * eta_xxxx * Ew * Ew * E2w )
						   / (SQR(omega_p) + SQR(-I*Omega) - I*Omega/tau)
					  * exp(- SQR(Omega/ Omega_BW) )
					  * exp(- SQR(inK[i] / K_BW) )
					  * 2. * sqrt(pi) / Omega_BW 
					  * 2. * sqrt(pi) / (K_BW)
					  * 2. * sqrt(pi) / (K_BW);
	}

	return Pol_array;
}



Complex
IntegrateKdKover4pi2_WRT_arbK( const valarray<double>& inK,
					const valarray<Complex>& inV,
					const double inK_cutoff)
{
	/*
 	 Integral over K for dK_vector (hence differential element is K dK)
	*/
	Complex r;

	CheckArray_(inV);

	assert(inV.size() == inK.size());
	
	r = 0.;
	unsigned long j = 1;
	
	for (j = 1;  j < inV.size() - 1; j++) {

		double Kj, Kj_1;
		
		Kj = inK[ j];
		Kj_1 = inK[ j - 1];

		double dK = Kj - Kj_1;
		if (dK == 0.) {
			cerr << "Zero dK\n";
		}

		if (Kj > inK_cutoff) {
		  cout << "K integral cut-off :" << Kj << " " << inK_cutoff << endl;
			
			double dK_co = inK_cutoff - Kj_1;
			complex_double Y_co = inV[ j - 1] + (inV[ j] - inV[ j - 1]) / dK * dK_co;
			
			r += (  inV[ j - 1] * Kj_1 + Y_co * inK_cutoff) * (dK_co) / (SQR(2. * pi)) ;
			break;
		} else {
			r += (  inV[ j - 1] * Kj_1 + inV[ j] * Kj) * (dK) / (SQR(2. * pi)) ;
		}
				
	}	
	return r;
}

Complex EpsilonAtOmegaForMaterial(double inOmega, unsigned long inMaterial)
{
	Complex e;
	
	switch (inMaterial) {
		case kInvalid:
			e = NAN;
			break;
		case kGaAs:
			e = epsilon( 3.6,1e-4);
			break;
	
		case kZnTe:
			e = epsilon( 3.2,1e-4);
			break;
			
		case kGaP:
			e = epsilon( 3.16,1e-4);
			break;
		case kAir:
			e = epsilon( 1,0);
			break;
		case kGaAsFrequencyDependent:
			{
				if (gGaAs_n_FrequencyDependent.size() == 0) {
					gGaAs_n_FrequencyDependent.attachFile("gaas_thz_index.dat");
				}
				if (gGaAs_k_FrequencyDependent.size() == 0) {
					gGaAs_k_FrequencyDependent.attachFile("gaas_thz_k.dat");
				}
				
				/*
				  The interpolating files are in THz in frequency, not Hz or 2 Pi f.
				*/
				double f = inOmega/ (2 * pi * 1e12);
				e = epsilon(gGaAs_n_FrequencyDependent.y(f), gGaAs_k_FrequencyDependent.y(f) );
			}
			break;
		case kZnTeFrequencyDependent:
			{
				if (gZnTe_n_FrequencyDependent.size() == 0) {
					gZnTe_n_FrequencyDependent.attachFile("znte_theo_index.dat");
				}
				if (gZnTe_k_FrequencyDependent.size() == 0) {
					gZnTe_k_FrequencyDependent.attachFile("znte_theo_k.dat");
				}
				
				/*
				  The interpolating files are in THz in frequency, not Hz or 2 Pi f.
				*/
				double f = inOmega/ (2 * pi * 1e12);
				e = epsilon(gZnTe_n_FrequencyDependent.y(f), gZnTe_k_FrequencyDependent.y(f) );
			}
			break;
		case kGaPFrequencyDependent:
			{
				if (gGaP_n_FrequencyDependent.size() == 0) {
					gGaP_n_FrequencyDependent.attachFile("gap_thz_index.dat");
				}
				if (gGaP_k_FrequencyDependent.size() == 0) {
					gGaP_k_FrequencyDependent.attachFile("gap_thz_k.dat");
				}
				
				/*
				  The interpolating files are in THz in frequency, not Hz or 2 Pi f.
				*/
				double f = inOmega/ (2 * pi * 1e12);
				e = epsilon(gGaP_n_FrequencyDependent.y(f), gGaP_k_FrequencyDependent.y(f) );
			}
			break;
		case kZnTeFrequencyDependentUpTo5THz:
			{
				if (gZnTe_n_FrequencyDependent.size() == 0) {
					gZnTe_n_FrequencyDependent.attachFile("znte_thz_grisch_index.txt");
				}
				if (gZnTe_k_FrequencyDependent.size() == 0) {
					gZnTe_k_FrequencyDependent.attachFile("znte_thz_grisch_k.txt");
				}
				
				/*
				  The interpolating files are in THz in frequency, not Hz or 2 Pi f.
				*/
				double f = inOmega/ (2 * pi * 1e12);
				e = Complex(epsilon(gZnTe_n_FrequencyDependent.y(f), gZnTe_k_FrequencyDependent.y(f) ));
			}
			break;
	}
	
	CheckValue_(e);
//	OUT << e << endl;
	return e;
}


Complex 
EOFilterAtOmega(double inOmega, Complex inEpsilon, double l, double ng_pump ) 
{
 	Complex filter, dk;
		
	if ( (inOmega == 0) || (index(inEpsilon) == ng_pump) ) {
      filter = 1;
	} else {
      dk = (index(inEpsilon) - ng_pump) * inOmega/ 3e8;
      
      filter = (exp(I*dk*l)-Complex(1.))/I/dk/Complex(l);
		
    }
    if (CheckValue_(filter)) {
    	cout << filter << " " << inEpsilon << " " << exp(I*dk*l) << " " << dk << endl;
    }

	return filter;
}


